package main_test

import (
	"context"
	"fmt"
	"sync"
	"sync/atomic"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
)

type Iterator interface {
	Iterate() (int, bool)
}

type ChannelIterator struct {
	ch chan int
}

func NewChannelIterator(data []int) Iterator {
	ch := make(chan int, len(data))
	for _, i := range data {
		ch <- i
	}
	close(ch)
	return &ChannelIterator{ch: ch}
}

func (iter *ChannelIterator) Iterate() (int, bool) {
	v, ok := <-iter.ch
	return v, ok
}

type SliceIterator struct {
	mu   sync.Mutex
	data []int
}

func NewSliceIterator(data []int) Iterator {
	clone := make([]int, len(data))
	copy(clone, data)
	return &SliceIterator{data: clone}
}

func (iter *SliceIterator) Iterate() (int, bool) {
	iter.mu.Lock()
	defer iter.mu.Unlock()
	if len(iter.data) == 0 {
		return 0, false
	}

	v := iter.data[0]
	iter.data = iter.data[1:]
	return v, true
}

type SliceIteratorNoLock struct {
	data []int
}

func NewSliceIteratorNoLock(data []int) Iterator {
	clone := make([]int, len(data))
	copy(clone, data)
	return &SliceIteratorNoLock{data: clone}
}

func (iter *SliceIteratorNoLock) Iterate() (int, bool) {
	if len(iter.data) == 0 {
		return 0, false
	}

	v := iter.data[0]
	iter.data = iter.data[1:]
	return v, true
}

func iterate(b *testing.B, iter Iterator, max int) int {
	count := 0
	for {
		v, ok := iter.Iterate()
		if !ok {
			break
		} else if max != -1 && count == max {
			break
		}
		count++
		_ = v
	}
	return count
}

func runIteration(b *testing.B, builder func([]int) Iterator) {
	data := []int{51, 2, 12, 30, 41}
	datalen := len(data)
	b.StopTimer()
	iter := builder(data)
	b.StartTimer()
	if count := iterate(b, iter, -1); count != datalen {
		b.Fatalf("expected %d elements, only got %d", datalen, count)
	}
}

func BenchmarkSliceIterator(b *testing.B) {
	for i := 0; i < b.N; i++ {
		runIteration(b, NewSliceIterator)
	}
}

func BenchmarkSliceIteratorNoLock(b *testing.B) {
	for i := 0; i < b.N; i++ {
		runIteration(b, NewSliceIteratorNoLock)
	}
}

func BenchmarkChannelIterator(b *testing.B) {
	for i := 0; i < b.N; i++ {
		runIteration(b, NewChannelIterator)
	}
}

func TestParallelSliceIterator(t *testing.T) {
	iter := NewSliceIterator([]int{51, 2, 12, 30, 41})
	for i := 0; i < 4; i++ {
		t.Run(fmt.Sprintf("goroutine %d", i), func(t *testing.T) {
			t.Parallel()
			for {
				v, ok := iter.Iterate()
				if !ok {
					return
				}
				t.Logf("%d", v)
				time.Sleep(500 * time.Millisecond)
			}
		})
	}
}

func TestParallelChannelIterator(t *testing.T) {
	iter := NewChannelIterator([]int{51, 2, 12, 30, 41})
	for i := 0; i < 4; i++ {
		t.Run(fmt.Sprintf("goroutine %d", i), func(t *testing.T) {
			t.Parallel()
			for {
				v, ok := iter.Iterate()
				if !ok {
					return
				}
				t.Logf("%d", v)
				time.Sleep(500 * time.Millisecond)
			}
		})
	}
}

type AtomicSequence struct {
	v int64
}

func NewAtomicSequence() Iterator {
	return &AtomicSequence{}
}

func (seq *AtomicSequence) Iterate() (int, bool) {
	v := atomic.AddInt64(&(seq.v), 1)
	return int(v), true
}

type InfiniteSequence struct {
	ch chan int
}

func NewInfiniteSequence(ctx context.Context) Iterator {
	ch := make(chan int, 1)
	go func(ctx context.Context, ch chan int) {
		defer close(ch)
		i := 1
		for {
			select {
			case <-ctx.Done():
				return
			case ch <- i:
				i++
			}
		}
	}(ctx, ch)
	return &InfiniteSequence{ch: ch}
}

func (seq *InfiniteSequence) Iterate() (int, bool) {
	v, ok := <-seq.ch
	return v, ok
}

func TestAtomicSequence(t *testing.T) {
	testSequence(t, NewAtomicSequence())
}

func TestInfiniteSequence(t *testing.T) {
	testSequence(t, NewInfiniteSequence(context.Background()))
}

func testSequence(t *testing.T, seq Iterator) {
	t.Helper()

	var seen = make(map[int]struct{})
	t.Run("Run AtomicSequence in parallel", func(t *testing.T) {
		var mu sync.Mutex
		for i := 0; i < 4; i++ {
			t.Run(fmt.Sprintf("goroutine %d", i), func(t *testing.T) {
				t.Parallel()
				for i := 1; i <= 100; i++ {
					v, ok := seq.Iterate()
					if !assert.True(t, ok, `seq.Iterate should be true`) {
						return
					}

					mu.Lock()
					if _, ok := seen[v]; !assert.False(t, ok, `same value (%d) seen multiple times`, v) {
						return
					}
					seen[v] = struct{}{}
					mu.Unlock()
				}
			})
		}
	})

	if !assert.Len(t, seen, 400, "sequence should produce 400 values") {
		return
	}
}

const maxSequence = 100

func BenchmarkAtomicSequence(b *testing.B) {
	b.StopTimer()
	seq := NewAtomicSequence()
	b.StartTimer()
	for i := 0; i < b.N; i++ {
		iterate(b, seq, maxSequence)
	}
}

func BenchmarkInfiniteSequence(b *testing.B) {
	b.StopTimer()
	seq := NewInfiniteSequence(context.Background())
	b.StartTimer()

	for i := 0; i < b.N; i++ {
		iterate(b, seq, maxSequence)
	}
}
